/*
 * $Id: sig_agp_bus.h,v 1.7 2009-01-27 17:44:19 potyra Exp $
 *
 * Copyright (C) 2008-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#ifndef __SIG_AGP_BUS_H_INCLUDED
#define __SIG_AGP_BUS_H_INCLUDED

#include <inttypes.h>

#include "sig_boolean.h"
#include "sig_boolean_or.h"
#include "sig_gen.h"

enum sig_agp_bus_main_cycle {
	/* Must be the same as in sig_host_bus.h! */
	SIG_AGP_BUS_C0R = 0,
	SIG_AGP_BUS_C0W,
	SIG_AGP_BUS_C1R,
	SIG_AGP_BUS_C1W,
	SIG_AGP_BUS_IOR,
	SIG_AGP_BUS_IOW,
	SIG_AGP_BUS_MR,
	SIG_AGP_BUS_MW,
};

struct sig_agp_bus_main_funcs {
	int (*type_addr)(void *s, unsigned int type, uint32_t addr);
	int (*read_data)(void *s, unsigned int bs, uint32_t *valp);
	int (*write_data)(void *s, unsigned int bs, uint32_t val);

	int (*c0r)(void *s, uint32_t addr, unsigned int bs, uint32_t *valp);
	int (*c0w)(void *s, uint32_t addr, unsigned int bs, uint32_t val);

	int (*c1r)(void *s, uint32_t addr, unsigned int bs, uint32_t *valp);
	int (*c1w)(void *s, uint32_t addr, unsigned int bs, uint32_t val);

	int (*ior)(void *s, uint32_t port, unsigned int bs, uint32_t *valp);
	int (*iow)(void *s, uint32_t port, unsigned int bs, uint32_t val);

	int (*ior_info)(void *s, uint32_t port, unsigned int bs,
			int (**cfp)(void *, uint32_t, unsigned int, uint32_t *),
			void **csp);
	int (*iow_info)(void *s, uint32_t port, unsigned int bs,
			int (**cfp)(void *, uint32_t, unsigned int, uint32_t),
			void **csp);

	void (*ior_info_flush)(void *s, uint32_t port, unsigned int bs);
	void (*iow_info_flush)(void *s, uint32_t port, unsigned int bs);

	int (*mr)(void *s, uint32_t addr, unsigned int bs, uint32_t *valp);
	int (*mw)(void *s, uint32_t addr, unsigned int bs, uint32_t val);

	int (*map)(void *s, unsigned long pa, unsigned int len,
			char **haddr_mr_p, char **haddr_mw_p);

	int (*unmap)(void *s, unsigned long pa, unsigned long len);
};

#define SIG_AGP_BUS_HASH_SIZE	(1 << 10)

struct sig_agp_bus_main {
	struct {
		void *s;
		const struct sig_agp_bus_main_funcs *f;
	} member[32];
	unsigned int member_count;

	struct sig_agp_bus_main_map {
		struct sig_agp_bus_main_map *lru_prev;
		struct sig_agp_bus_main_map *lru_next;

		struct sig_agp_bus_main_map *hash_prev;
		struct sig_agp_bus_main_map *hash_next;

		uint32_t addr;
		uint32_t *va_mr;
		uint32_t *va_mw;
	} map[256];
	struct sig_agp_bus_main_map *map_lru_first;
	struct sig_agp_bus_main_map *map_lru_last;
	struct sig_agp_bus_main_map *map_hash_first[SIG_AGP_BUS_HASH_SIZE];
	struct sig_agp_bus_main_map *map_hash_last[SIG_AGP_BUS_HASH_SIZE];

	struct sig_agp_bus_main_ior {
		struct sig_agp_bus_main_ior *lru_prev;
		struct sig_agp_bus_main_ior *lru_next;

		struct sig_agp_bus_main_ior *hash_prev;
		struct sig_agp_bus_main_ior *hash_next;

		uint32_t port;
		unsigned int bs;
		int (*f)(void *s,
				uint32_t port, unsigned int bs, uint32_t *valp);
		void *s;
	} ior[256];
	struct sig_agp_bus_main_ior *ior_lru_first;
	struct sig_agp_bus_main_ior *ior_lru_last;
	struct sig_agp_bus_main_ior *ior_hash_first[SIG_AGP_BUS_HASH_SIZE];
	struct sig_agp_bus_main_ior *ior_hash_last[SIG_AGP_BUS_HASH_SIZE];

	struct sig_agp_bus_main_iow {
		struct sig_agp_bus_main_iow *lru_prev;
		struct sig_agp_bus_main_iow *lru_next;

		struct sig_agp_bus_main_iow *hash_prev;
		struct sig_agp_bus_main_iow *hash_next;

		uint32_t port;
		unsigned int bs;
		int (*f)(void *s,
				uint32_t port, unsigned int bs, uint32_t val);
		void *s;
	} iow[256];
	struct sig_agp_bus_main_iow *iow_lru_first;
	struct sig_agp_bus_main_iow *iow_lru_last;
	struct sig_agp_bus_main_iow *iow_hash_first[SIG_AGP_BUS_HASH_SIZE];
	struct sig_agp_bus_main_iow *iow_hash_last[SIG_AGP_BUS_HASH_SIZE];
};

struct sig_agp_bus_main_merge {
	struct sig_agp_bus_main *s0;
	struct sig_agp_bus_main *s1;
};

extern int
sig_agp_bus_main_type_addr(struct sig_agp_bus_main *b, void *s,
		unsigned int type, uint32_t addr);
extern int
sig_agp_bus_main_read_data(struct sig_agp_bus_main *b, void *s,
		unsigned int bs, uint32_t *valp);
extern int
sig_agp_bus_main_write_data(struct sig_agp_bus_main *b, void *s,
		unsigned int bs, uint32_t val);

extern int
sig_agp_bus_main_c0r(struct sig_agp_bus_main *b, void *s,
		uint32_t addr, unsigned int bs, uint32_t *valp);
extern int
sig_agp_bus_main_c0w(struct sig_agp_bus_main *b, void *s,
		uint32_t addr, unsigned int bs, uint32_t val);

extern int
sig_agp_bus_c1r(struct sig_agp_bus_main *b, void *s,
		uint32_t addr, unsigned int bs, uint32_t *valp);
extern int
sig_agp_bus_c1w(struct sig_agp_bus_main *b, void *s,
		uint32_t addr, unsigned int bs, uint32_t val);

extern int
sig_agp_bus_ior(struct sig_agp_bus_main *b, void *s,
		uint32_t port, unsigned int bs, uint32_t *valp);
extern int
sig_agp_bus_iow(struct sig_agp_bus_main *b, void *s,
		uint32_t port, unsigned int bs, uint32_t val);

extern int
sig_agp_bus_ior_info(struct sig_agp_bus_main *b, void *s,
		uint32_t port, unsigned int bs,
		int (**cf)(void *, uint32_t, unsigned int, uint32_t *),
		void **cs);
extern int
sig_agp_bus_iow_info(struct sig_agp_bus_main *b, void *s,
		uint32_t port, unsigned int bs,
		int (**cf)(void *, uint32_t, unsigned int, uint32_t),
		void **cs);

extern void
sig_agp_bus_ior_info_flush(struct sig_agp_bus_main *b, void *s,
		uint32_t port, unsigned int bs);
extern void
sig_agp_bus_iow_info_flush(struct sig_agp_bus_main *b, void *s,
		uint32_t port, unsigned int bs);

extern int
sig_agp_bus_mr(struct sig_agp_bus_main *b, void *s,
		uint32_t addr, unsigned int bs, uint32_t *valp);
extern int
sig_agp_bus_mw(struct sig_agp_bus_main *b, void *s,
		uint32_t addr, unsigned int bs, uint32_t val);
extern int
sig_agp_bus_map(struct sig_agp_bus_main *b, void *s,
		unsigned long pa, unsigned int len,
		char **haddr_mr_p, char **haddr_mw_p);

extern int
sig_agp_bus_unmap(struct sig_agp_bus_main *b,
		void *s, unsigned long pa, unsigned long len);

extern void
sig_agp_bus_main_connect(struct sig_agp_bus_main *b,
		void *s, const struct sig_agp_bus_main_funcs *f);

extern struct sig_agp_bus_main *
sig_agp_bus_main_init(const char *name, int nr);

extern struct sig_agp_bus_main_merge *
sig_agp_bus_main_merge(
	struct sig_agp_bus_main *s0,
	struct sig_agp_bus_main *s1
);
extern void
sig_agp_bus_main_split(struct sig_agp_bus_main_merge *m);

extern void
sig_agp_bus_main_create(const char *name, int nr);
extern void
sig_agp_bus_main_destroy(const char *name, int nr);

struct sig_agp_bus {
	enum sig_gen_type type;
	struct sig_boolean *p3_3V;
	struct sig_boolean *p5V;
	struct sig_boolean *p12V;
	struct sig_boolean *n_reset;
	struct sig_agp_bus_main *main;
	struct sig_boolean_or *intA;
	struct sig_boolean_or *intB;
};

struct sig_agp_bus *
sig_agp_bus_init(const char *name, unsigned int nr);

extern void
sig_agp_bus_create(const char *name, unsigned int nr);
extern void
sig_agp_bus_destroy(const char *name, unsigned int nr);

#endif /* __SIG_AGP_BUS_H_INCLUDED */
