/* $Id: sig_boolean_or.c,v 1.13 2009-01-27 17:06:41 potyra Exp $ 
 *
 * Copyright (C) 2008-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fixme.h"

#include "sig_boolean_or.h"

void
sig_boolean_or_flush(struct sig_boolean_or *b)
{
	unsigned int nr;
	unsigned int i;

	for (i = 0; i < b->in_count; i++) {
		void (*func)(void *, unsigned int);
		unsigned int new_val;

		new_val = 0;

		for (nr = 0; nr < b->out_count; nr++) {
			if (b->out[nr].s == b->in[i].s) continue;

			new_val |= b->out[nr].out;
		}

		func = b->in[i].f->set_ext;
		if (func
		 && b->in[i].in != new_val) {
			b->in[i].in = new_val;
			func(b->in[i].s, new_val);
		}

		for (nr = 0; nr < b->out_count; nr++) {
			if (b->out[nr].s != b->in[i].s) continue;

			new_val |= b->out[nr].out;
		}

		func = b->in[i].f->set;
		if (func
		 && b->in[i].in != new_val) {
			b->in[i].in = new_val;
			func(b->in[i].s, new_val);
		}
	}
}

void
sig_boolean_or_set(struct sig_boolean_or *b, void *s, unsigned int val)
{
	unsigned int nr;

again:	;
	for (nr = 0; ; nr++) {
		if (nr == b->out_count) {
			fprintf(stderr, "WARNING: late registration of boolean_or driver of %p!\n", s);
			sig_boolean_or_connect_out(b, s, 0);
			goto again;
		}
		assert(nr < b->out_count);
		if (b->out[nr].s == s) {
			b->out[nr].out = val;
			break;
		}
	}

	sig_boolean_or_flush(b);
}

void
sig_boolean_or_connect_in(
	struct sig_boolean_or *b,
	void *s,
	const struct sig_boolean_or_funcs *f
)
{
	assert(b);
	assert(b->in_count < sizeof(b->in) / sizeof(b->in[0]));
	assert(s);
	assert(f);
	assert(f->set || f->set_ext);
	assert(! f->set || ! f->set_ext);

	b->in[b->in_count].s = s;
	b->in[b->in_count].f = f;
	b->in[b->in_count].in = 0;
	b->in_count++;

	sig_boolean_or_flush(b);
}

void
sig_boolean_or_connect_out(
	struct sig_boolean_or *b,
	void *s,
	unsigned int val
)
{
	assert(b);
	assert(b->out_count < sizeof(b->out) / sizeof(b->out[0]));
	// assert(s);
	assert(val == 0 || val == 1);

	if (! s) {
		fprintf(stderr, "WARNING: %s: s is NULL pointer!\n",
				__FUNCTION__);
	}

	b->out[b->out_count].s = s;
	b->out[b->out_count].out = val;
	b->out_count++;

	sig_boolean_or_flush(b);
}

struct sig_boolean_or *
sig_boolean_or_init(const char *name, int nr)
{
	struct sig_boolean_or *b;

	b = malloc(sizeof(*b));
	assert(b);
	memset(b, 0, sizeof(*b));

	b->type = SIG_GEN_BOOLEAN_OR;
	b->in_count = 0;
	b->out_count = 0;

	return b;
}

static void
sig_boolean_or_s0_set(void *_f, unsigned int val)
{
	struct sig_boolean_or_merge *f = (struct sig_boolean_or_merge *) _f;

	sig_boolean_or_set(f->s1, f, val);
}

static void
sig_boolean_or_s1_set(void *_f, unsigned int val)
{
	struct sig_boolean_or_merge *f = (struct sig_boolean_or_merge *) _f;

	sig_boolean_or_set(f->s0, f, val);
}

struct sig_boolean_or_merge *
sig_boolean_or_merge(struct sig_boolean_or *s0, struct sig_boolean_or *s1)
{
	static struct sig_boolean_or_funcs s0_funcs = {
		.set_ext = sig_boolean_or_s0_set,
	};
	static struct sig_boolean_or_funcs s1_funcs = {
		.set_ext = sig_boolean_or_s1_set,
	};
	struct sig_boolean_or_merge *m;

	m = malloc(sizeof(*m));
	assert(m);

	/* Out */
	m->s0 = s0;
	sig_boolean_or_connect_out(s0, m, 0);
	m->s1 = s1;
	sig_boolean_or_connect_out(s1, m, 0);

	/* In */
	sig_boolean_or_connect_in(s0, m, &s0_funcs);
	sig_boolean_or_connect_in(s1, m, &s1_funcs);

	return m;
}

void
sig_boolean_or_split(struct sig_boolean_or_merge *m)
{
	fixme();
}

void
sig_boolean_or_create(const char *name, int nr)
{
}

void
sig_boolean_or_destroy(const char *name, int nr)
{
}
