/*
 * $Id: isa_gen_ne2000_gui_gtk.c,v 1.19 2009-03-06 14:21:23 sand Exp $
 *
 * Copyright (C) 2003-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "glue-gui-gtk.h"
#include "glue-gui-gtk-led.h"
#include "glue-gui-gtk-fi.h"

#include "isa_gen_ne2000_gui_gtk.h"

#define COMP "isa_gen_ne2000"

struct cpssp {
	GtkWidget *busy_led;
	GtkWidget *fi_dialog;

	struct sig_fault *fault_recv_loss;
	struct sig_fault *fault_send_loss;
};

static struct cpssp *iface[100];

/*
 * Simulator Callbacks
 */
static void
gui_network_opt_busy_led_set(void *_cpssp, unsigned int val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	gui_gtk_led_set(GUI_GTK_LED(cpssp->busy_led), val);
	gui_gtk_flush();
}

/*
 * GUI Callbacks
 */
static void
isa_gen_ne2000_gui_recv_loss_event(
	GtkWidget *w,
	gpointer _fault,
	gpointer _cpssp
)
{
	struct fault {
		unsigned int state;
		const char *loc0[4];
		const char *loc1[4];
	} *fault = (struct fault *) _fault;
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_fault_set(cpssp->fault_recv_loss, cpssp, 0ULL, 0ULL, fault->state);

	free(fault);
}

static void
isa_gen_ne2000_gui_send_loss_event(
	GtkWidget *w,
	gpointer _fault,
	gpointer _cpssp
)
{
	struct fault {
		unsigned int state;
		const char *loc0[4];
		const char *loc1[4];
	} *fault = (struct fault *) _fault;
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_fault_set(cpssp->fault_send_loss, cpssp, 0ULL, 0ULL, fault->state);

	free(fault);
}

void
isa_gen_ne2000_gui_gtk_init(
	unsigned int nr,
	struct sig_isa_bus *port_isa,
	struct sig_eth *port_eth,
	struct sig_boolean *port_opt_busy_led,
	struct sig_fault *fault_recv_loss,
	struct sig_fault *fault_send_loss
)
{
	static const struct sig_boolean_funcs opt_busy_led_funcs = {
		.set = gui_network_opt_busy_led_set,
	};
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	/* Out */
	cpssp->fault_recv_loss = fault_recv_loss;
	cpssp->fault_send_loss = fault_send_loss;

	/* In */
	sig_boolean_connect_in(port_opt_busy_led, cpssp, &opt_busy_led_funcs);
}

void
isa_gen_ne2000_gui_gtk_create(
	unsigned int page,
	unsigned int nr,
	const char *name,
	const char *irq,
	const char *ioaddr,
	const char *mac
)
{
	struct cpssp *cpssp;
	GtkWidget *vbox;
	GtkWidget *w;

	cpssp = malloc(sizeof(*cpssp));
	assert(cpssp);

	vbox = gtk_vbox_new(FALSE, 1);

	/* Busy LED */
	cpssp->busy_led = gui_gtk_led_new("Busy");
	gtk_widget_show(cpssp->busy_led);
	gtk_box_pack_start(GTK_BOX(vbox), cpssp->busy_led, FALSE, FALSE, 1);

	/* FI Dialog */
	cpssp->fi_dialog = gtk_vbox_new(FALSE, 1);
	
	w = gui_gtk_fi_new("Receive Loss", "Percentage");
	g_signal_connect(G_OBJECT(w), "fi-change",
			G_CALLBACK(isa_gen_ne2000_gui_recv_loss_event),
			cpssp);
	gtk_widget_show(w);
	gtk_box_pack_start(GTK_BOX(cpssp->fi_dialog), w, FALSE, FALSE, 1);

	w = gui_gtk_fi_new("Send Loss", "Percentage");
	g_signal_connect(G_OBJECT(w), "fi-change",
			G_CALLBACK(isa_gen_ne2000_gui_send_loss_event),
			cpssp);
	gtk_widget_show(w);
	gtk_box_pack_start(GTK_BOX(cpssp->fi_dialog), w, FALSE, FALSE, 1);

	gui_gtk_comp_add(page, COMP, nr, name, vbox,
			FALSE, FALSE, cpssp->fi_dialog);

	iface[nr] = cpssp;
}

void
isa_gen_ne2000_gui_gtk_destroy(unsigned int nr)
{
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	free(cpssp);

	iface[nr] = NULL;
}
