/* $Id: mb_ibm_at_io_decode.c,v 1.5 2009-01-28 12:59:20 potyra Exp $ 
 *
 * Copyright (C) 2008-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include <assert.h>
#include <stdio.h>
#include "fixme.h"

#include "glue-shm.h"

#include "mb_ibm_at_io_decode.h"

#define COMP	"mb_ibm_at_io_decode"

struct cpssp {
	struct sig_cs *port_dma0_cs;
	struct sig_cs *port_dma1_cs;
	struct sig_cs *port_dmapage_cs;
	struct sig_cs *port_kbd_cs;
	struct sig_cs *port_ppi_cs;
	struct sig_cs *port_pit_cs;
	struct sig_cs *port_pic0_cs;
	struct sig_cs *port_pic1_cs;
	struct sig_cs *port_rtc_cs;
};

static int
mb_ibm_at_io_decode_inb(void *_cpssp, unsigned char *valp, unsigned short port)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	/*
	 * IBM AT
	 * 1-28
	 */
	switch (port) {
	case 0x0000 ... 0x001f:
		/* DMA controller 1 */
		return sig_cs_readb(cpssp->port_dma0_cs, cpssp,
				valp, port & 0x1f);
	case 0x0020 ... 0x003f:
		/* Interrupt controller 1 */
		return sig_cs_readb(cpssp->port_pic0_cs, cpssp,
				valp, port & 0x1f);
	case 0x0040 ... 0x005f:
		/* Timer */
		return sig_cs_readb(cpssp->port_pit_cs, cpssp,
				valp, port & 0x1f);
	case 0x0060:
	case 0x0064:
		/* 8042 (Keyboard) */
		return sig_cs_readb(cpssp->port_kbd_cs, cpssp,
				valp, port & 0xf);
	case 0x0061:
	case 0x0063:
		/* PPI */
		return sig_cs_readb(cpssp->port_ppi_cs, cpssp,
				valp, (port >> 1) & 1);
	case 0x0070 ... 0x007f:
		/* Real-time clock, NMI */
		return sig_cs_readb(cpssp->port_rtc_cs, cpssp,
				valp, port & 0xf);
	case 0x0080 ... 0x009f:
		/* DMA page register */
		return sig_cs_readb(cpssp->port_dmapage_cs, cpssp,
				valp, port);
	case 0x00a0 ... 0x00bf:
		/* Interrupt controller 2 */
		return sig_cs_readb(cpssp->port_pic1_cs, cpssp,
				valp, port & 0x1f);
	case 0x00c0 ... 0x00df:
		/* DMA controller 2 */
		return sig_cs_readb(cpssp->port_dma1_cs, cpssp,
				valp, (port >> 1) & 0x1f);
	case 0x00f0 ... 0x00ff:
		/* Math Coprocessor */
		fixme();
		return 0;
	default:
		return -1;
	}
}

static int
mb_ibm_at_io_decode_outb(void *_cpssp, unsigned char val, unsigned short port)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	/*
	 * IBM AT
	 * 1-28
	 */
	switch (port) {
	case 0x0000 ... 0x001f:
		/* DMA controller 1 */
		return sig_cs_writeb(cpssp->port_dma0_cs, cpssp,
				val, port & 0x1f);
	case 0x0020 ... 0x003f:
		/* Interrupt controller 1 */
		return sig_cs_writeb(cpssp->port_pic0_cs, cpssp,
				val, port & 0x1f);
	case 0x0040 ... 0x005f:
		/* Timer */
		return sig_cs_writeb(cpssp->port_pit_cs, cpssp,
				val, port & 0x1f);
	case 0x0060:
	case 0x0064:
		/* 8042 (Keyboard) */
		return sig_cs_writeb(cpssp->port_kbd_cs, cpssp,
				val, port & 0xf);
	case 0x0061:
	case 0x0063:
		/* PPI */
		return sig_cs_writeb(cpssp->port_ppi_cs, cpssp,
				val, (port >> 1) & 1);
	case 0x0070 ... 0x007f:
		/* Real-time clock, NMI */
		return sig_cs_writeb(cpssp->port_rtc_cs, cpssp,
				val, port & 0xf);
	case 0x0080 ... 0x009f:
		/* DMA page register */
		return sig_cs_writeb(cpssp->port_dmapage_cs, cpssp,
				val, port & 0x1f);
	case 0x00a0 ... 0x00bf:
		/* Interrupt controller 2 */
		return sig_cs_writeb(cpssp->port_pic1_cs, cpssp,
				val, port & 0x1f);
	case 0x00c0 ... 0x00df:
		/* DMA controller 2 */
		return sig_cs_writeb(cpssp->port_dma1_cs, cpssp,
				val, (port >> 1) & 0x1f);
	case 0x00f0 ... 0x00ff: /* Math Coprocessor */
		fixme();
		return 0;
	default:
		return -1;
	}
}

void
mb_ibm_at_io_decode_init(
	unsigned int nr,
	struct sig_isa_bus_main *port_bus,
	struct sig_cs *port_dma0_cs,
	struct sig_cs *port_dma1_cs,
	struct sig_cs *port_dmapage_cs,
	struct sig_cs *port_kbd_cs,
	struct sig_cs *port_ppi_cs,
	struct sig_cs *port_pit_cs,
	struct sig_cs *port_pic0_cs,
	struct sig_cs *port_pic1_cs,
	struct sig_cs *port_rtc_cs
)
{
	static const struct sig_isa_bus_main_funcs bus_funcs = {
		.inb = mb_ibm_at_io_decode_inb,
		.outb = mb_ibm_at_io_decode_outb,
	};
	struct cpssp *cpssp;

	cpssp = shm_map(COMP, nr, sizeof(*cpssp), 0);

	cpssp->port_dma0_cs = port_dma0_cs;
	cpssp->port_dma1_cs = port_dma1_cs;
	cpssp->port_dmapage_cs = port_dmapage_cs;
	cpssp->port_kbd_cs = port_kbd_cs;
	cpssp->port_ppi_cs = port_ppi_cs;
	cpssp->port_pit_cs = port_pit_cs;
	cpssp->port_pic0_cs = port_pic0_cs;
	cpssp->port_pic1_cs = port_pic1_cs;
	cpssp->port_rtc_cs = port_rtc_cs;

	sig_isa_bus_main_connect(port_bus, cpssp, &bus_funcs);
}

unsigned int
mb_ibm_at_io_decode_create(void)
{
	static unsigned int nr = 0;

	shm_create(COMP, nr, sizeof(struct cpssp));

	return nr++;
}

void
mb_ibm_at_io_decode_destroy(unsigned int nr)
{
	shm_destroy(COMP, nr);
}
