/* $Id: mb_ibm_at_reset.c,v 1.4 2009-01-28 12:59:20 potyra Exp $ 
 *
 * Copyright (C) 2008-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include <assert.h>
#include <stdio.h>
#include "fixme.h"

#include "glue-main.h"
#include "glue-shm.h"

#include "mb_ibm_at_reset.h"

#define COMP	"mb_ibm_at_reset"

struct cpssp {
	/* Ports */
	struct sig_std_logic *port_n_button;
	unsigned int state_n_button;
	unsigned int state_n_kbc_reset;
	struct sig_boolean *port_n_reset;
	
	/* State */
	unsigned int reset_triggered;
};

static void
mb_ibm_at_reset_reset_update(struct cpssp *cpssp)
{
	unsigned int reset;

	reset = cpssp->reset_triggered
		| ! cpssp->state_n_button
		| ! cpssp->state_n_kbc_reset;

	sig_boolean_set(cpssp->port_n_reset, cpssp, ! reset);
}

static void
mb_ibm_at_reset_event(void *_cpssp)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	cpssp->reset_triggered = 0;
	mb_ibm_at_reset_reset_update(cpssp);
}

static void
mb_ibm_at_reset_power_set(void *_cpssp, unsigned int val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_std_logic_set(cpssp->port_n_button, cpssp,
			val ? SIG_STD_LOGIC_H : SIG_STD_LOGIC_L);

	if (val) {
		cpssp->reset_triggered = 1;
		mb_ibm_at_reset_reset_update(cpssp);
		time_call_after(TIME_HZ / 10, mb_ibm_at_reset_event, cpssp);
	} else {
		/* Remove timer - FIXME */
	}
}

static void
mb_ibm_at_reset_n_button_set(void *_cpssp, unsigned int n_val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	switch (n_val) {
	case SIG_STD_LOGIC_0:
	case SIG_STD_LOGIC_L: n_val = 0; break;
	case SIG_STD_LOGIC_1:
	case SIG_STD_LOGIC_H: n_val = 1; break;
	default: /* FIXME */ return;
	}

	cpssp->state_n_button = n_val;
	mb_ibm_at_reset_reset_update(cpssp);
}

static void
mb_ibm_at_reset_n_kbc_reset_set(void *_cpssp, unsigned int n_val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;
	
	cpssp->state_n_kbc_reset = n_val;
	mb_ibm_at_reset_reset_update(cpssp);
}

void
mb_ibm_at_reset_init(
	unsigned int nr,
	struct sig_boolean *port_power,
	struct sig_std_logic *port_button_hash_,
	struct sig_boolean *port_kbc_reset_hash_,
	struct sig_boolean *port_reset_hash_
)
{
	static const struct sig_boolean_funcs power_funcs = {
		.set = mb_ibm_at_reset_power_set,
	};
	static const struct sig_std_logic_funcs n_button_funcs = {
		.set = mb_ibm_at_reset_n_button_set,
	};
	static const struct sig_boolean_funcs n_kbc_reset_funcs = {
		.set = mb_ibm_at_reset_n_kbc_reset_set,
	};
	struct cpssp *cpssp;

	cpssp = shm_map(COMP, nr, sizeof(*cpssp), 0);

	/* Out */
	cpssp->port_n_button = port_button_hash_;
	cpssp->state_n_button = 0;
	sig_std_logic_connect_out(port_button_hash_, cpssp, SIG_STD_LOGIC_L);

	cpssp->port_n_reset = port_reset_hash_;
	sig_boolean_connect_out(port_reset_hash_, cpssp, 0);

	/* In */
	sig_boolean_connect_in(port_power, cpssp, &power_funcs);
	sig_std_logic_connect_in(port_button_hash_, cpssp, &n_button_funcs);
	sig_boolean_connect_in(port_kbc_reset_hash_, cpssp, &n_kbc_reset_funcs);
}

unsigned int
mb_ibm_at_reset_create(void)
{
	static unsigned int nr = 0;

	shm_create(COMP, nr, sizeof(struct cpssp));

	return nr++;
}

void
mb_ibm_at_reset_destroy(unsigned int nr)
{
	shm_destroy(COMP, nr);
}
