/*
 *  $Id: monitor_gui_null.c,v 1.2 2009-03-07 09:17:19 sand Exp $
 *
 * Copyright (C) 2007-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "monitor_gui_null.h"
#include "glue-gui-null-monitor.h"

#define COMP "monitor"

struct cpssp {
	void * gui;
};

static struct cpssp *iface[100];

/*
 * Simulator Callbacks
 */
static void
monitor_gui_pixel_set(
	void *_cpssp,
	unsigned int x,
	unsigned int y,
	uint8_t r,
	uint8_t g,
	uint8_t b
)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	gui_null_monitor_pixel_set(cpssp->gui, x, y, r, g, b);
}

static void
monitor_gui_size_set(void *_cpssp, unsigned int width, unsigned int height)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	gui_null_monitor_size_set(cpssp->gui, width, height);
}

static void
monitor_gui_sync(void *_cpssp)
{
#if 0
	struct cpssp *cpssp = (struct cpssp *) _cpssp;
#endif
}

static void
monitor_gui_screenshot_set(void *_cpssp, int val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	if (val == 0) return;

	gui_null_monitor_screenshot(cpssp->gui, val);
}

/*
 * GUI Callbacks
 */

void
monitor_gui_null_init(
	unsigned int nr,
	struct sig_boolean *port_power_230v,
	struct sig_vga *port_vga,
	struct sig_opt_rgb *port_opt_video,
	struct sig_integer *port_screenshot
)
{
	static const struct sig_opt_rgb_funcs opt_video_funcs = {
		.pixel_set = monitor_gui_pixel_set,
		.size_set = monitor_gui_size_set,
		.sync = monitor_gui_sync,
	};
	static const struct sig_integer_funcs screenshot_funcs = {
		.set = monitor_gui_screenshot_set,
	};
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	/* In */
	sig_opt_rgb_connect(port_opt_video, cpssp, &opt_video_funcs);
	sig_integer_connect_in(port_screenshot, cpssp, &screenshot_funcs);
}

void
monitor_gui_null_create(unsigned int page, unsigned int nr, const char *name)
{
	struct cpssp *cpssp;

	cpssp = malloc(sizeof(*cpssp));
	assert(cpssp);

	cpssp->gui = gui_null_monitor_create(1600,1200);

	iface[nr] = cpssp;
}

void
monitor_gui_null_destroy(unsigned int nr)
{
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	gui_null_monitor_destroy(cpssp->gui);

	free(cpssp);

	iface[nr] = NULL;
}
