/*
 * $Id: power_supply_at_gui_gtk.c,v 1.12 2009-03-06 13:02:19 sand Exp $
 *
 * Copyright (C) 2003-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "glue-gui-gtk.h"
#include "glue-gui-gtk-switch.h"

#include "power_supply_at_gui_gtk.h"

#define COMP "power_supply_at"

struct cpssp {
	GtkWidget *widget;

	struct sig_boolean *port_mech_power_switch;
};

static struct cpssp *iface[100];

/*
 * Simulator Callbacks
 */
static void
power_supply_at_mech_power_switch_set(void *_cpssp, unsigned int val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	gui_gtk_switch_set(GUI_GTK_SWITCH(cpssp->widget), (gboolean) val);
	gui_gtk_flush();
}

/*
 * GUI Callbacks
 */
static void
power_supply_at_mech_power_switched_on_event(GtkWidget *w, gpointer _cpssp)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_boolean_set(cpssp->port_mech_power_switch, cpssp, 1);
}

static void
power_supply_at_mech_power_switched_off_event(GtkWidget *w, gpointer _cpssp)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_boolean_set(cpssp->port_mech_power_switch, cpssp, 0);
}

void
power_supply_at_gui_gtk_init(
	unsigned int nr,
	struct sig_boolean *port_power_230v,
	struct sig_boolean *port_mech_power_switch,
	struct sig_power_board_at *port_power_board,
	struct sig_power_device *port_power_dev0,
	struct sig_power_device *port_power_dev1,
	struct sig_power_device *port_power_dev2,
	struct sig_power_device *port_power_dev3,
	struct sig_power_device *port_power_dev4,
	struct sig_power_device *port_power_dev5,
	struct sig_power_device *port_power_dev6,
	struct sig_power_device *port_power_dev7
)
{
	static const struct sig_boolean_funcs mech_power_switch_funcs = {
		.set = power_supply_at_mech_power_switch_set,
	};
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	/* Out */
	cpssp->port_mech_power_switch = port_mech_power_switch;
	sig_boolean_connect_out(port_mech_power_switch, cpssp, 0);

	/* In */
	sig_boolean_connect_in(port_mech_power_switch, cpssp,
			&mech_power_switch_funcs);
}

void
power_supply_at_gui_gtk_create(unsigned int page, unsigned int nr, const char *name)
{
	struct cpssp *cpssp;
	GtkWidget *vbox;

	cpssp = malloc(sizeof(*cpssp));
	assert(cpssp);

	vbox = gtk_vbox_new(FALSE, 1);

	cpssp->widget = gui_gtk_switch_new();
	g_signal_connect(G_OBJECT(cpssp->widget), "switched-on",
			G_CALLBACK(power_supply_at_mech_power_switched_on_event),
			cpssp);
	g_signal_connect(G_OBJECT(cpssp->widget), "switched-off",
			G_CALLBACK(power_supply_at_mech_power_switched_off_event),
			cpssp);
	gtk_widget_show(cpssp->widget);
	gtk_box_pack_end(GTK_BOX(vbox), cpssp->widget, FALSE, FALSE, 1);

	gtk_widget_show(vbox);
	gui_gtk_comp_add(page, COMP, nr, name, vbox, FALSE, FALSE, NULL);

	iface[nr] = cpssp;
}

void
power_supply_at_gui_gtk_destroy(unsigned int nr)
{
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	free(cpssp);

	iface[nr] = NULL;
}
