/*
 * $Id: scsi_gen_cdrom_gui_gtk.c,v 1.19 2009-03-26 14:54:05 vrsieh Exp $
 *
 * Copyright (C) 2003-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "glue-gui-gtk.h"
#include "glue-gui-gtk-change.h"
#include "glue-gui-gtk-fi.h"

#include "scsi_gen_cdrom_gui_gtk.h"

#define COMP "scsi_gen_cdrom"

struct cpssp {
	struct sig_string *port_change;

	GtkWidget *gui_change;
	GtkWidget *gui_fi;
};

static struct cpssp *iface[100];

/*
 * Simulator Callbacks
 */

static void
scsi_gen_cdrom_gui_change(void *_cpssp, const char *path)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	gui_gtk_change_set(GUI_GTK_CHANGE(cpssp->gui_change), path);
}

/*
 * GUI Callbacks
 */
static void
scsi_gen_cdrom_insert_event(GtkWidget *w, const char *string, gpointer _cpssp)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_string_set(cpssp->port_change, cpssp, string);
}

static void
scsi_gen_cdrom_remove_event(GtkWidget *w, gpointer _cpssp)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	sig_string_set(cpssp->port_change, cpssp, "");
}

static void
scsi_gen_cdrom_gui_drive_fault_event(
	GtkWidget *w,
	gpointer _fault,
	gpointer _cpssp
)
{
	struct fault {
		unsigned int state;
		const char *loc0[4];
		const char *loc1[4];
	} *fault = (struct fault *) _fault;
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	fprintf(stderr, "%s %p %u\n", __FUNCTION__, cpssp, fault->state);

	free(fault);
}

void
scsi_gen_cdrom_gui_gtk_init(
	unsigned int nr,
	struct sig_power_device *port_power,
	struct sig_scsi_bus *port_scsi,
	struct sig_magneto_optical *port_media,
	struct sig_string *port_change
)
{
	static const struct sig_string_funcs change_funcs = {
		.set = scsi_gen_cdrom_gui_change,
	};
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	/* Out */
	cpssp->port_change = port_change;

	/* In */
	sig_string_connect(port_change, cpssp, &change_funcs);
}

void
scsi_gen_cdrom_gui_gtk_create(
	unsigned int page,
	unsigned int nr,
	const char *name,
	const char *scsi_id_str
)
{
	struct cpssp *cpssp;
	GtkWidget *vbox;

	cpssp = malloc(sizeof(*cpssp));
	assert(cpssp);

	vbox = gtk_vbox_new(FALSE, 1);

	/* Change Dialog */
	cpssp->gui_change = gui_gtk_change_new("CD");
	GTK_WIDGET_UNSET_FLAGS(cpssp->gui_change, GTK_CAN_FOCUS);
	g_signal_connect(G_OBJECT(cpssp->gui_change), "change-inserted",
			G_CALLBACK(scsi_gen_cdrom_insert_event), cpssp);
	g_signal_connect(G_OBJECT(cpssp->gui_change), "change-removed",
			G_CALLBACK(scsi_gen_cdrom_remove_event), cpssp);
	gtk_widget_show(cpssp->gui_change);
	gtk_box_pack_end(GTK_BOX(vbox), cpssp->gui_change, FALSE, FALSE, 1);

	/* FI Dialog */
	cpssp->gui_fi = gui_gtk_fi_new("Drive Fault", "Boolean");
	g_signal_connect(G_OBJECT(cpssp->gui_fi), "fi-change",
			G_CALLBACK(scsi_gen_cdrom_gui_drive_fault_event),
			cpssp);

	gtk_widget_show(vbox);
	gtk_widget_show(cpssp->gui_fi);
	gui_gtk_comp_add(page, COMP, nr, name, vbox, FALSE, FALSE, cpssp->gui_fi);

	iface[nr] = cpssp;
}

void
scsi_gen_cdrom_gui_gtk_destroy(unsigned int nr)
{
	struct cpssp *cpssp;

	cpssp = iface[nr];
	assert(cpssp);

	free(cpssp);

	iface[nr] = NULL;
}
